#include "my_random.hpp"
#include "stereo_vision.hpp"


#include <stdio.h>
#include <iostream>
#include <fstream>
#include <opencv2/core.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>

const double threshold = 3.;
const int ransac_max_iteration = 5000;
const double ransac_confidence = 0.99;
const std::string image_src_name = "data/0000.jpg";
const std::string image_dest_name = "data/0001.jpg";
const std::string matching_name = "data/matching.txt";
const std::string calibration_matrix_name = "data/K.txt";

int main(int argc, char** argv)
{
	// Setting the random seed to get random results in each run.
	my_random::init_random();

	// Read images
	cv::Mat image_src = cv::imread(image_src_name);
	cv::Mat image_dest = cv::imread(image_dest_name);

	// Define features detector
	std::cout << matching_name << std::endl;

	std::vector<cv::Point2d> src_points, dest_points;
	stereo_vision::read_matching_file(matching_name, src_points, dest_points);

	// Show the correspondences
	cv::Mat draw_matches_output;
	stereo_vision::draw_correspondences(image_src, src_points, image_dest, dest_points, draw_matches_output);
	cv::imwrite("data/all_correspondences.jpg", draw_matches_output);

	// Normalize points
	std::cout << "Normalize points..." << std::endl;
	std::vector<cv::Point2d> normalized_src_points, normalized_dest_points;
	cv::Mat T_src, T_dest;
	stereo_vision::normalize_points(src_points, normalized_src_points, T_src);
	stereo_vision::normalize_points(dest_points, normalized_dest_points, T_dest);

	// Find fundamental matrix
	std::cout << "Finding fundamental matrix..." << std::endl;
	cv::Mat F_normalized;
	std::vector<int> inliers;

	// Since we use the normalized coordinates, we have to normalize the ransac threshold
	// We scale the threshold with the average scales of the src and dest images
	// (another solution is to find the inliers according to original coordinates after denormalize the fundamental matrix)
	double normalized_threshold = threshold * ((T_src.at<double>(0, 0) + T_dest.at<double>(0, 0)) / 2.);

	stereo_vision::ransac_fundamental_matrix_8points_method(
		normalized_src_points,
		normalized_dest_points,
		F_normalized,
		inliers,
		ransac_max_iteration,      // max iterations
		normalized_threshold,      // ransac threshold
		ransac_confidence          // ransac confidence
	);

	// Local optimization
	stereo_vision::get_fundamental_matrix_LSQ(
		normalized_src_points,
		normalized_dest_points,
		inliers,
		F_normalized
	);

	std::cout << "F_normalized: " << std::endl
		<< F_normalized << std::endl;
	cv::Mat F = T_dest.t() * F_normalized * T_src;
	std::cout << "F:" << std::endl
		<< F << std::endl;

	// Draw inliers correspondences
	stereo_vision::draw_correspondences(image_src, src_points, image_dest, dest_points, inliers, draw_matches_output);
	cv::imwrite("data/inliers_correspondences.jpg", draw_matches_output);

	// Calibration matrix
	cv::Mat K = stereo_vision::read_camera_matrix(calibration_matrix_name);
	// Essential matrix
	cv::Mat E = K.t() * F * K;
	std::cout << "E:" << std::endl
		<< E << std::endl;

	cv::Mat img1 = image_src;
	cv::Mat img2 = image_dest;
	std::vector<cv::Point3d> lines1, lines2;

	cv::computeCorrespondEpilines(src_points, 1, F, lines2);
	//cv::computeCorrespondEpilines(dest_points, 2, F, lines1);

	std::cout << std::endl;
	for (int i = 0; i < 10; ++i) {
		double R = my_random::random_number() * 255, B = my_random::random_number() * 255, G = my_random::random_number() * 255;
		int idx = my_random::random_number() * (src_points.size()-1);

		cv::circle(img1, src_points[idx], 15, { R, G, B }, cv::FILLED);
		cv::circle(img2, dest_points[idx], 15, { R, G, B }, cv::FILLED);

		cv::Mat_<double> p(3, 1);
		p(0, 0) = src_points[idx].x;
		p(1, 0) = src_points[idx].y;
		p(2, 0) = 1.0;
		cv::Mat l = F * p;
		double aa = l.at<double>(0, 0);
		double bb = l.at<double>(1, 0);
		double cc = l.at<double>(2, 0);

		double nu = cv::sqrt(aa * aa + bb * bb);
		aa /= nu;
		bb /= nu;
		cc /= nu;

		double a = lines2[idx].x;
		double b = lines2[idx].y;
		double c = lines2[idx].z;
		cv::Point2d p1 = { 0.0, -c / b };
		double x = img2.cols;
		cv::Point2d p2 = { x,  (-a * x - c) / b };
		cv::line(img2, p1, p2, { R,G,B }, 7);
	}


	cv::imwrite("1.jpg", img1);
	cv::imwrite("2.jpg", img2);

	//
	cv::Mat R1, R2, t;
	stereo_vision::decompose_essential_matrix(
		E,
		R1,
		R2,
		t
	);


	cv::Mat P1, P2, R, correct_t;
	stereo_vision::get_correct_rotation_translation(
		src_points,
		dest_points,
		inliers,
		K,
		R1,
		R2,
		t,
		R,
		correct_t,
		P1,
		P2
	);

	std::cout << "P1: " << std::endl
		<< P1 << std::endl;
	std::cout << "P2: " << std::endl
		<< P2 << std::endl;
	std::cout << "R: " << std::endl
		<< R << std::endl;
	std::cout << "t: " << std::endl
		<< correct_t << std::endl;

	return 0;
}